<?php

declare(strict_types=1);

namespace Dedi\SyliusMigrationPlugin\Service\Sql;

use Dedi\SyliusMigrationPlugin\Definition\TableMappingDefinition;
use Dedi\SyliusMigrationPlugin\Dto\OrderMappingDefinition;
use Doctrine\DBAL\Query\QueryBuilder;
use LogicException;

class Reader extends AbstractSqlProcessor implements ReaderInterface
{
    public function read(TableMappingDefinition $tableMappingDefinition, int $offset, int $limit): array
    {
        $sourceQueryBuilder = $this->resolveQueryBuilder($tableMappingDefinition);
        $sql = $this->resolveSqlStatement($tableMappingDefinition, $sourceQueryBuilder, $offset, $limit);
        $parameters = $this->resolveParameters($tableMappingDefinition, $sourceQueryBuilder);
        $this->logger->debug($tableMappingDefinition, $sql);

        return $this->connection->executeQuery($sql, $parameters)->fetchAllAssociative();
    }

    public function getResultCount(TableMappingDefinition $tableMappingDefinition, int $limit): int
    {
        $sourceQueryBuilder = $this->resolveCountQueryBuilder($tableMappingDefinition);
        $sql = sprintf('SELECT COUNT(*) FROM (%s) a', $this->resolveSqlStatement($tableMappingDefinition, $sourceQueryBuilder, 0, $limit));
        $parameters = $this->resolveParameters($tableMappingDefinition, $sourceQueryBuilder);
        $this->logger->debug($tableMappingDefinition, $sql);

        $count = $this->connection->executeQuery($sql, $parameters)->fetchOne();

        return false !== $count ? (int) $count : 0;
    }

    private function addSelectPart(TableMappingDefinition $tableMappingDefinition, QueryBuilder $queryBuilder): void
    {
        foreach ($tableMappingDefinition->getColumnMappingDefinitions() as $columnMappingDefinition) {
            if (null !== $columnMappingDefinition->getSourceColumnName()) {
                $queryBuilder
                    ->addSelect($columnMappingDefinition->getSourceColumnName())
                ;
            }

            if (null !== $columnMappingDefinition->getExpr()) {
                if (null === $columnMappingDefinition->getAlias()) {
                    throw new LogicException('Expr configuration needs an Alias');
                }
                $queryBuilder
                    ->addSelect(sprintf('%s AS %s', $columnMappingDefinition->getExpr(), $columnMappingDefinition->getAlias()))
                ;
            }
        }
    }

    private function addSearchPart(TableMappingDefinition $tableMappingDefinition, QueryBuilder $queryBuilder): void
    {
        if ($tableMappingDefinition->hasFilterMapping()) {
            foreach ($tableMappingDefinition->getFilterMapping() as $filterMappingDefinition) {
                $pattern = $this->resolveSearchType($filterMappingDefinition->getType());

                $queryBuilder
                    ->andWhere(sprintf($pattern, $filterMappingDefinition->getFilterColumn(), $filterMappingDefinition->getFilterColumn()))
                    ->setParameter($filterMappingDefinition->getFilterColumn(), $filterMappingDefinition->getValue())
                ;
            }
        }
    }

    private function addOrderPart(TableMappingDefinition $tableMappingDefinition, QueryBuilder $queryBuilder): void
    {
        $orderMappingDefinition = $tableMappingDefinition->getOrderMappingDefinition();

        if ($orderMappingDefinition instanceof OrderMappingDefinition) {
            $queryBuilder->orderBy($orderMappingDefinition->getColumn(), $orderMappingDefinition->getSort());
        } else {
            $queryBuilder->orderBy($tableMappingDefinition->getSourcePrimaryKey(), 'DESC');
        }
    }

    private function createQueryBuilder(TableMappingDefinition $tableMappingDefinition): QueryBuilder
    {
        $sourceQueryBuilder = $this->connection->createQueryBuilder();
        $sourceQueryBuilder
            ->from($tableMappingDefinition->getSourceTableName())
            ->addSelect($tableMappingDefinition->getSourcePrimaryKey())
        ;

        return $sourceQueryBuilder;
    }

    private function resolveParameters(TableMappingDefinition $tableMappingDefinition, ?QueryBuilder $queryBuilder): array
    {
        if ('' !== $tableMappingDefinition->getSql()) {
            return [];
        }

        if (null !== $queryBuilder) {
            return $queryBuilder->getParameters();
        }

        throw new LogicException(sprintf('Could not resolve query parameters for table %s', $tableMappingDefinition->getTargetTableName()));
    }

    private function resolveQueryBuilder(TableMappingDefinition $tableMappingDefinition): ?QueryBuilder
    {
        if ('' !== $tableMappingDefinition->getSql()) {
            return null;
        }
        $sourceQueryBuilder = $this->createQueryBuilder($tableMappingDefinition);
        $this->addSelectPart($tableMappingDefinition, $sourceQueryBuilder);
        $this->addSearchPart($tableMappingDefinition, $sourceQueryBuilder);
        $this->addOrderPart($tableMappingDefinition, $sourceQueryBuilder);

        return $sourceQueryBuilder;
    }

    private function resolveCountQueryBuilder(TableMappingDefinition $tableMappingDefinition): ?QueryBuilder
    {
        if ('' !== $tableMappingDefinition->getSql()) {
            return null;
        }

        $sourceQueryBuilder = $this->createQueryBuilder($tableMappingDefinition);
        $this->addSearchPart($tableMappingDefinition, $sourceQueryBuilder);
        $this->addOrderPart($tableMappingDefinition, $sourceQueryBuilder);

        return $sourceQueryBuilder;
    }

    private function resolveSqlStatement(TableMappingDefinition $tableMappingDefinition, ?QueryBuilder $queryBuilder, int $offset, int $limit): string
    {
        if ('' !== $tableMappingDefinition->getSql()) {
            return $this->addLimitPartToSql($tableMappingDefinition->getSql(), $offset, $limit);
        }

        if (null !== $queryBuilder) {
            $this->addLimitPartToQueryBuilder($queryBuilder, $offset, $limit);

            return $queryBuilder->getSQL();
        }

        throw new LogicException(sprintf('Could not resolve a runnable query for table %s', $tableMappingDefinition->getTargetTableName()));
    }

    private function resolveSearchType(string $type): string
    {
        switch ($type) {
            case 'comparison':
                return '%s = :%s';
            case 'not_comparison':
                return '%s != :%s';
            case 'like':
                return '%s LIKE :%s';
            case 'is_null':
                return '%s IS NULL';
            case 'in':
                return '%s IN (:%s)';
            default:
                throw new LogicException('Unsupported type %s', $type);
        }
    }

    private function addLimitPartToSql(string $query, int $offset, int $limit): string
    {
        if ($limit === -1) {
            return $query;
        }

        return sprintf(trim($query) . ' LIMIT %s, %s', $offset, $limit);
    }

    private function addLimitPartToQueryBuilder(QueryBuilder $queryBuilder, $offset, $limit): void
    {
        if ($limit === -1) {
            return;
        }

        $queryBuilder
            ->setFirstResult($offset)
            ->setMaxResults($limit)
        ;
    }
}
